from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional

from casanova import TabularRecord


@dataclass
class BuzzsumoArticle(TabularRecord):
    buzzsumo_id: Optional[str]
    twitter_shares: Optional[int]
    facebook_shares: Optional[int]
    url: Optional[str]
    updated_at: Optional[datetime]
    facebook_likes: Optional[int]
    total_facebook_shares: Optional[int]
    title: Optional[str]
    published_date: Optional[datetime]
    pinterest_shares: Optional[int]
    og_url: Optional[str]
    article_amplifiers: Optional[str]
    domain_name: Optional[str]
    author_name: Optional[str]
    num_words: Optional[int]
    subdomain: Optional[str]
    twitter_user_id: Optional[str]
    video_length: Optional[int]
    language: Optional[str]
    alexa_rank: Optional[int]
    facebook_comments: Optional[int]
    num_linking_domains: Optional[int]
    youtube_views: Optional[int]
    wow_count: Optional[int]
    love_count: Optional[int]
    haha_count: Optional[int]
    sad_count: Optional[int]
    angry_count: Optional[int]
    youtube_likes: Optional[int]
    youtube_dislikes: Optional[int]
    youtube_comments: Optional[int]
    youtube_trending_score: Optional[str]
    tiktok_download_count: Optional[int]
    tiktok_play_count: Optional[int]
    tiktok_share_count: Optional[int]
    tiktok_comment_count: Optional[int]
    tiktok_digg_count: Optional[int]
    tiktok_author: Optional[str]
    tiktok_music_id: Optional[str]
    tiktok_music_author: Optional[str]
    tiktok_music_title: Optional[str]
    total_shares: Optional[int]
    thumbnail: Optional[str]
    total_reddit_engagements: Optional[int]
    evergreen_score: Optional[float]
    evergreen_score2: Optional[float]
    article_types: Optional[List[str]]
    author_image: Optional[str]
    ecommerce: Optional[str]
    is_general_article: bool
    is_how_to_article: bool
    is_infographic: bool
    is_interview: bool
    is_list: bool
    is_newsletter: bool
    is_podcast: bool
    is_presentation: bool
    is_press_release: bool
    is_review: bool
    is_video: bool
    is_what_post: bool
    is_why_post: bool

    @classmethod
    def from_payload(cls, payload: Dict) -> "BuzzsumoArticle":
        return cls(
            buzzsumo_id=payload.get("id"),
            twitter_shares=payload.get("twitter_shares"),
            facebook_shares=payload.get("facebook_shares"),
            url=payload.get("url"),
            updated_at=datetime.fromtimestamp(payload["updated_at"]) or None,
            facebook_likes=payload.get("facebook_likes"),
            total_facebook_shares=payload.get("total_facebook_shares"),
            title=payload.get("title"),
            published_date=datetime.fromtimestamp(payload["updated_at"]) or None,
            pinterest_shares=payload.get("pinterest_shares"),
            og_url=payload.get("og_url"),
            article_amplifiers=payload.get("article_amplifiers"),
            domain_name=payload.get("domain_name"),
            author_name=payload.get("author_name"),
            num_words=payload.get("num_words"),
            subdomain=payload.get("subdomain"),
            twitter_user_id=payload.get("twitter_user_id"),
            video_length=payload.get("video_length"),
            language=payload.get("language"),
            alexa_rank=payload.get("alexa_rank"),
            facebook_comments=payload.get("facebook_comments"),
            num_linking_domains=payload.get("num_linking_domains"),
            youtube_views=payload.get("youtube_views"),
            wow_count=payload.get("wow_count"),
            love_count=payload.get("love_count"),
            haha_count=payload.get("haha_count"),
            sad_count=payload.get("sad_count"),
            angry_count=payload.get("angry_count"),
            youtube_likes=payload.get("youtube_likes"),
            youtube_dislikes=payload.get("youtube_dislikes"),
            youtube_comments=payload.get("youtube_comments"),
            youtube_trending_score=payload.get("youtube_trending_score"),
            tiktok_download_count=payload.get("tiktok_download_count"),
            tiktok_play_count=payload.get("tiktok_play_count"),
            tiktok_share_count=payload.get("tiktok_share_count"),
            tiktok_comment_count=payload.get("tiktok_comment_count"),
            tiktok_digg_count=payload.get("tiktok_digg_count"),
            tiktok_author=payload.get("tiktok_author"),
            tiktok_music_id=payload.get("tiktok_music_id"),
            tiktok_music_author=payload.get("tiktok_music_author"),
            tiktok_music_title=payload.get("tiktok_music_title"),
            total_shares=payload.get("total_shares"),
            thumbnail=payload.get("thumbnail"),
            total_reddit_engagements=payload.get("total_reddit_engagements"),
            evergreen_score=payload.get("evergreen_score"),
            evergreen_score2=payload.get("evergreen_score2"),
            article_types=payload.get("article_types"),
            author_image=payload.get("author_image"),
            ecommerce=payload.get("ecommerce"),
            is_general_article=bool(payload.get("general_article")),
            is_how_to_article=bool(payload.get("how_to_article")),
            is_infographic=bool(payload.get("infographic")),
            is_interview=bool(payload.get("interview")),
            is_list=bool(payload.get("list")),
            is_newsletter=bool(payload.get("newsletter")),
            is_podcast=bool(payload.get("podcast")),
            is_presentation=bool(payload.get("presentation")),
            is_press_release=bool(payload.get("press_release")),
            is_review=bool(payload.get("review")),
            is_video=bool(payload.get("video")),
            is_what_post=bool(payload.get("what_post")),
            is_why_post=bool(payload.get("why_post")),
        )



from typing import List, Optional

from urllib3 import Timeout

from minet.cli.argparse import (
    command,
    FolderStrategyType,
    BooleanAction,
    ExtractionSelectionAction,
    FOLDER_STRATEGY_DOCUMENTATION,
)
from minet.cli.exceptions import InvalidArgumentsError

# TODO: lazyloading issue
from minet.constants import DEFAULT_THROTTLE

# TODO: negative int forbiding type
CRAWL_ARGUMENTS = {
    "output_dir": {
        "flags": ["-O", "--output-dir"],
        "help": "Output directory.",
        "default": "crawl",
    },
    "resume": {
        "flag": "--resume",
        "help": "Whether to resume an interrupted crawl.",
        "action": "store_true",
    },
    "max_depth": {
        "flag": "--max-depth",
        "help": "Maximum depth for the crawl.",
        "type": int,
    },
    "throttle": {
        "flag": "--throttle",
        "help": "Time to wait - in seconds - between 2 calls to the same domain.",
        "type": float,
        "default": DEFAULT_THROTTLE,
    },
    "domain_parallelism": {
        "flag": "--domain-parallelism",
        "help": "Max number of urls per domain to hit at the same time.",
        "type": int,
        "default": 1,
    },
    "threads": {
        "flags": ["-t", "--threads"],
        "help": "Number of threads to use. Will default to a conservative number, based on the number of available cores. Feel free to increase.",
        "type": int,
    },
    "compress_on_disk": {
        "flags": ["-z", "--compress-on-disk"],
        "help": "Whether to compress the downloaded files when saving files on disk.",
        "action": "store_true",
    },
    "compress_transfer": {
        "flag": "--compress-transfer",
        "help": 'Whether to send a "Accept-Encoding" header asking for a compressed response. Usually better for bandwidth but at the cost of more CPU work.',
        "action": "store_true",
    },
    "write_files": {
        "flags": ["-w", "--write-files"],
        "help": "Whether to write downloaded files on disk in order to save them for later.",
        "action": "store_true",
    },
    "write_data": {
        "flags": ["-d", "--write-data", "-D", "--dont-write-data"],
        "help": "Whether to write scraped/extracted data on disk.",
        "action": BooleanAction,
        "default": True,
    },
    "folder_strategy": {
        "flag": "--folder-strategy",
        "help": "Name of the strategy to be used to dispatch the retrieved files into folders to alleviate issues on some filesystems when a folder contains too much files. Note that this will be applied on top of --filename-template. All of the strategies are described at the end of this help.",
        "default": "flat",
        "type": FolderStrategyType(),
    },
    "format": {
        "flags": ["-f", "--format"],
        "help": "Serialization format for scraped/extracted data.",
        "choices": ("csv", "jsonl", "ndjson"),
        "default": "csv",
    },
    "verbose": {
        "flags": ["-v", "--verbose"],
        "help": "Whether to print information about crawl results.",
        "action": "store_true",
    },
    "visit_urls_only_once": {
        "flags": ["-u", "--visit-urls-only-once"],
        "help": "Whether to ensure that any url will only be visited once.",
        "action": "store_true",
    },
    "normalized_url_cache": {
        "flags": ["-n", "--normalized-url-cache"],
        "help": "Whether to normalize url cache used to assess if some url was already visited.",
        "action": "store_true",
    },
    "insecure": {
        "flags": ["-k", "--insecure"],
        "help": "Whether to allow ssl errors when performing requests or not.",
        "action": "store_true",
    },
    "spoof_user_agent": {
        "flag": "--spoof-user-agent",
        "help": 'Whether to use a plausible random "User-Agent" header when making requests.',
        "action": "store_true",
    },
    "processes": {
        "flags": ["-p", "--processes"],
        "help": "Number of processes for the crawler process pool.",
        "type": int,
    },
    "connect_timeout": {
        "flag": "--connect-timeout",
        "help": "Maxium socket connection time to host.",
        "type": float,
        "default": 5,
    },
    "timeout": {
        "flag": "--timeout",
        "help": "Maximum time - in seconds - to spend for each request before triggering a timeout.",
        "type": float,
        "default": 30,
    },
    "retries": {
        "flag": "--retries",
        "help": "Number of times to retry on timeout & common network-related issues.",
        "type": int,
        "default": 0,
    },
    "stateful_redirects": {
        "flag": "--stateful-redirects",
        "help": "Whether to keep a cookie jar while redirecting and allowing self redirections that track state.",
        "action": "store_true",
    },
    "pycurl": {
        "flag": "--pycurl",
        "help": "Whether to use the pycurl library to perform the calls.",
        "action": "store_true",
    },
    "sqlar": {
        "flag": "--sqlar",
        "help": "Whether to write files into a single-file sqlite archive rather than as individual files on the disk.",
        "action": "store_true",
    },
}

CRAWL_MODULE_ARGUMENTS = [
    {
        "flags": ["-m", "--module"],
        "help": "Python module to import to use as spider, spiders or crawler factory. Suffix it with `:` to give actual target within module e.g. `package.module:spider`.",
    },
    {
        "flag": "--factory",
        "help": "Whether crawl target is a crawler factory function.",
        "action": "store_true",
    },
]


def delete(o, k):
    try:
        del o[k]
    except KeyError:
        pass


def crawl_command(
    name: str,
    package: str,
    title: Optional[str] = None,
    description: Optional[str] = None,
    epilog: Optional[str] = None,
    arguments: Optional[List] = None,
    accept_input: bool = True,
    accept_module: bool = False,
    resolve=None,
    unique: bool = False,
    url_cache: bool = True,
    max_depth: bool = True,
    throttle: bool = True,
    default_throttle: Optional[float] = None,
    domain_parallelism: bool = True,
    threads: bool = True,
    write_files: bool = False,
    write_data: bool = True,
    factory: bool = False,
    default_folder_strategy: Optional[str] = None,
    force_folder_strategy: Optional[str] = None,
    default_output_dir: Optional[str] = None,
    default_retries: Optional[int] = None,
    force_spoof_user_agent: Optional[bool] = None,
    force_stateful_redirects: Optional[bool] = None,
    timeout: bool = True,
    default_timeout: Optional[float] = None,
    default_connect_timeout: Optional[float] = None,
):
    arguments_dict = CRAWL_ARGUMENTS.copy()

    def set_default_value(k, v):
        arguments_dict[k] = {
            **arguments_dict[k],
            "default": v,
        }

    # NOTE: missing a lot in the resolve here
    if unique:
        delete(arguments_dict, "visit_urls_only_once")

    if not url_cache:
        delete(arguments_dict, "visit_urls_only_once")
        delete(arguments_dict, "normalized_url_cache")

    if not max_depth:
        del arguments_dict["max_depth"]

    if not throttle:
        del arguments_dict["throttle"]

    if default_throttle is not None:
        set_default_value("throttle", default_throttle)

    if not threads:
        del arguments_dict["threads"]

    if not domain_parallelism:
        del arguments_dict["domain_parallelism"]

    if write_files:
        del arguments_dict["write_files"]

    if not write_data:
        del arguments_dict["write_data"]
        del arguments_dict["format"]

    if default_folder_strategy is not None:
        set_default_value("folder_strategy", default_folder_strategy)

    if default_output_dir is not None:
        set_default_value("output_dir", default_output_dir)

    if default_retries is not None:
        set_default_value("retries", default_retries)

    if not timeout:
        del arguments_dict["timeout"]
        del arguments_dict["connect_timeout"]

    else:
        if default_timeout is not None:
            set_default_value("timeout", default_timeout)

        if default_connect_timeout is not None:
            set_default_value("connect_timeout", default_connect_timeout)

    if force_folder_strategy is not None:
        del arguments_dict["folder_strategy"]

    if force_spoof_user_agent is not None:
        del arguments_dict["spoof_user_agent"]

    if force_stateful_redirects is not None:
        del arguments_dict["stateful_redirects"]

    arguments = (arguments or []) + list(arguments_dict.values())

    if accept_module:
        arguments.extend(CRAWL_MODULE_ARGUMENTS)

    additional_kwargs = {}

    if accept_input:
        additional_kwargs["variadic_input"] = {
            "dummy_column": "url",
            "item_label": "start url",
            "optional": True,
            "no_help": True,
        }

        arguments.append(
            {
                "flag": "--input-spider",
                "help": "Name of the spider that will process jobs given using -i/--input.",
            }
        )

    epilog = (
        (epilog or "")
        + """
        Notes regarding sqlite storage temporary files:

        If you spawn a persistent crawler, some of its state will be kept using
        specialized sqlite databases on disk.

        Be aware that sqlite sometimes need to perform cleanup operations that rely
        on temporary files that will be written in a folder configured as the
        `SQLITE_TMPDIR` env variable. Be sure to set this variable when running this
        command, especially when the crawler files will be stored on a different disk
        and if writing in the host's `/tmp` directory can be risky due to insufficient
        storage available.

        For more info, check out: https://www.sqlite.org/tempfiles.html
        """
    )

    if "folder_strategy" in arguments_dict:
        # NOTE: text indentation IS important
        epilog = f"""
        --folder-strategy options:

        {FOLDER_STRATEGY_DOCUMENTATION}

        """ + (
            epilog or ""
        )

    def wrapped_resolve(cli_args):
        if unique:
            cli_args.visit_urls_only_once = True

        if write_files:
            cli_args.write_files = True

        if not write_data:
            cli_args.write_data = False

        if not accept_module:
            cli_args.module = None
            cli_args.factory = False

        if factory:
            cli_args.factory = True

        if force_folder_strategy is not None:
            cli_args.folder_strategy = force_folder_strategy

        if force_spoof_user_agent is not None:
            cli_args.spoof_user_agent = force_spoof_user_agent

        if force_stateful_redirects is not None:
            cli_args.stateful_redirects = force_stateful_redirects

        if resolve is not None:
            resolve(cli_args)

        if timeout:
            cli_args.timeout = Timeout(
                connect=cli_args.connect_timeout, total=cli_args.timeout
            )

        # --sqlar disables --compress-on-disk
        if cli_args.sqlar:
            cli_args.compress_on_disk = False

    return command(
        name,
        package,
        title=title,
        description=description,
        epilog=epilog,
        arguments=arguments,
        no_output=True,
        resolve=wrapped_resolve,
        **additional_kwargs,
    )


CRAWL_COMMAND = crawl_command(
    "crawl",
    "minet.cli.crawl.crawl",
    title="Minet Crawl Command",
    description="""
        Run a crawl using a minet crawler or spiders defined
        in a python module.
    """,
    epilog=f"""
        Examples:

        . Using the most basic crawler following HTML links:
            $ minet crawl url -i urls.csv --max-depth 1

        . Crawling using the `process` function from the `crawl` module:
            $ minet crawl -m crawl:process -O crawl-data
    """,
    accept_module=True,
)


def check_focus_crawl_arguments(cli_args):
    if not cli_args.content_filter and not cli_args.url_filter:
        raise InvalidArgumentsError(
            [
                "At least one filter is required, either for URLs or content.",
                "To do so, use one of the following flags:",
                "   -C/--content-filter",
                "   -U/--url-filter",
            ]
        )

    if cli_args.extraction_fields and not cli_args.extract:
        raise InvalidArgumentsError(
            "Custom extraction fields can't be used if the flag extract is not present."
        )


FOCUS_CRAWL_COMMAND = crawl_command(
    "focus-crawl",
    "minet.cli.crawl.focus",
    title="Minet Focus Crawl Command",
    description="""
        Minet crawl feature with the possibility
        to use regular expressions to filter content.

        Regex are not case sensitive, but
        accents sensitive.

        Regex must be written between simple quotes.
    """,
    epilog="""
        Examples:

        . Running a simple crawler:
            $ minet focus-crawl url -i urls.csv --content-filter '(?:assembl[ée]e nationale|s[ée]nat)' -O ./result
    """,
    resolve=check_focus_crawl_arguments,
    unique=True,
    arguments=[
        {
            "flags": ["-C", "--content-filter"],
            "help": "Regex used to filter fetched content.",
            "default": None,
        },
        {
            "flags": ["-U", "--url-filter"],
            "help": "Regex used to filter URLs added to crawler's queue.",
            "default": None,
        },
        {
            "flag": "--invert-content-match",
            "help": "Flag to turn the content filter into an exclusion rule.",
            "action": "store_true",
        },
        {
            "flag": "--invert-url-match",
            "help": "Flag to turn the url filter into an exclusion rule.",
            "action": "store_true",
        },
        {
            "flag": "--extract",
            "help": "Perform regex match on extracted text content instead of html content using the Trafilatura library.",
            "action": "store_true",
        },
        {
            "flag": "--irrelevant-continue",
            "help": "Continue exploration whether met content is relevant or not.",
            "action": "store_true",
        },
        {
            "flag": "--only-html",
            "help": "Add URLs to the crawler queue only if they seem to lead to a HTML content.",
            "action": "store_true",
        },
        {
            "flag": "--extraction-fields",
            "help": "Fields of the trafilatura extraction you want to apply the content filter on, separated using commas. It must be used with the flag `--extract`.",
            "action": ExtractionSelectionAction,
            "default": None,
        },
    ],
)



# =============================================================================
# Minet Rich Console
# =============================================================================
#
# Rich console instance used by minet CLI.
#
from typing import Optional

from datetime import datetime
from rich.console import Console
from rich.theme import Theme
from rich.table import Table

from minet.utils import message_flatmap

NOW_TIME_FORMAT = r"%H:%M:%S"
NOW_DATETIME_FORMAT = r"%Y-%m-%d %H:%M:%S"

MINET_COLORS = {
    "info": "blue",
    "error": "#d42a20",
    "warning": "#fac22b",
    "cream": "#fcf3d9",
    "dark": "#02080b",
    "success": "green",
}

MINET_COLOR_BACKGROUNDS = {}

for name, style in MINET_COLORS.items():
    MINET_COLOR_BACKGROUNDS[name + "_background"] = "on " + style

MINET_STYLES = {
    **MINET_COLORS,
    **MINET_COLOR_BACKGROUNDS,
    # Progress bar
    "bar.complete": MINET_COLORS["info"],
    "bar.finished": MINET_COLORS["success"],
    "progress.percentage": "",
    "progress.elapsed": "",
}
MINET_THEME = Theme(MINET_STYLES)


class MinetConsole(Console):
    def vprint(self, *messages, style: Optional[str] = None):
        txt = message_flatmap(*messages)
        self.print(txt, style=style)

    def logh(self, header: str, *messages, style=None, header_style="log.time") -> None:
        txt = message_flatmap(*messages)

        table = Table.grid(padding=(0, 1), expand=True)

        table.add_column(style=header_style)
        table.add_column(ratio=1, overflow="fold", style=style)

        table.add_row(header, txt)

        self.print(table)

    def info(self, *messages):
        self.logh("info", *messages, header_style="info")

    def error(self, *messages):
        self.logh("error", *messages, header_style="error")

    def warning(self, *messages):
        self.logh("warning", *messages, header_style="warning")

    def log_with_time(self, *messages, full=False, style=None):
        now = datetime.now().strftime(NOW_DATETIME_FORMAT if full else NOW_TIME_FORMAT)
        now = "[%s]" % now

        self.logh(now, *messages, style=style)


console = MinetConsole(theme=MINET_THEME, stderr=True, highlight=False)

__all__ = ["console"]

if __name__ == "__main__":
    for color in MINET_COLORS:
        console.print(color, style=color, end=" ")
    console.print("\n")

    console.log_with_time("hello")



from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.loading_bar import LoadingBar

from minet.facebook.emulated_scraper import FacebookEmulatedScraper
from minet.facebook.exceptions import FacebookInvalidTargetError
from minet.facebook.types import FacebookComment


@with_enricher_and_loading_bar(
    headers=FacebookComment,
    title="Scraping comments",
    unit="posts",
    nested=True,
    sub_unit="comments",
)
def action(cli_args, enricher, loading_bar: LoadingBar):
    with FacebookEmulatedScraper(headless=not cli_args.headful) as scraper:
        for i, row, url in enricher.enumerate_cells(
            cli_args.column, with_rows=True, start=1
        ):
            with loading_bar.step(url):
                try:
                    comments = scraper.scrape_comments(url)
                except FacebookInvalidTargetError:
                    loading_bar.print(
                        "Given url (line %i) is probably not a Facebook group post: %s"
                        % (i, url)
                    )
                    continue

                for comment in comments:
                    enricher.writerow(row, comment)

                loading_bar.nested_advance(len(comments))



# =============================================================================
# Minet Twitter Friends CLI Action
# =============================================================================
#
# Logic of the `tw followers` action.
#
from minet.cli.twitter.utils import make_twitter_action

REPORT_HEADERS = ["follower_id"]

action = make_twitter_action(method_name="followers", csv_headers=REPORT_HEADERS)



# =============================================================================
# Minet Hyphe Declare CLI Action
# =============================================================================
#
# Logic of the `hyphe declare` action.
#
import casanova
from urllib.parse import unquote

from minet.cli.loading_bar import LoadingBar
from minet.cli.exceptions import InvalidArgumentsError
from minet.cli.hyphe.utils import with_hyphe_fatal_errors
from minet.hyphe import HypheAPIClient
from minet.hyphe.exceptions import HypheRequestFailError


def extract_tags(row, tag_pos_list):
    tags = {}

    for name, pos in tag_pos_list:
        cell_value = row[pos]

        if not cell_value:
            continue

        # TODO: hyphe tag serialization is not bijective
        tags[name] = cell_value.split("|")

    return {"USER": tags}


@with_hyphe_fatal_errors
def action(cli_args):
    reader = casanova.reader(cli_args.webentities, total=cli_args.total)
    headers = reader.headers

    name_pos = headers.get("NAME")
    homepage_pos = headers.get("HOME PAGE")
    prefixes_pos = headers.get("PREFIXES AS LRU")
    status_pos = headers.get("STATUS")
    startpages_pos = headers.get("START PAGES")

    tag_pos_list = [
        (name.split("(TAGS)", 1)[0].strip(), pos)
        for name, pos in headers
        if name.endswith("(TAGS)")
    ]

    if (
        name_pos is None
        or prefixes_pos is None
        or homepage_pos is None
        or status_pos is None
        or startpages_pos is None
    ):
        raise InvalidArgumentsError(
            "input csv file is not a valid hyphe webentities export"
        )

    client = HypheAPIClient(cli_args.url)
    corpus = client.corpus(cli_args.corpus, password=cli_args.password)

    corpus.ensure_is_started()

    with LoadingBar(
        title="Declaring web entities", unit="webentities", total=reader.total
    ) as loading_bar:
        for row in reader:
            with loading_bar.step():
                startpages_cell = row[startpages_pos]

                name = row[name_pos]

                # TODO: bug on hyphe's side
                if "h:localhost|" in row[prefixes_pos]:
                    loading_bar.print("Dropping %s because of localhost issue" % name)
                    continue

                homepage = row[homepage_pos]
                prefixes = row[prefixes_pos].split(" ")
                startpages = startpages_cell.split(" ") if startpages_cell else []
                startpages = list(
                    set(startpages)
                )  # TODO: dedupe because of hyphe issues
                tags = extract_tags(row, tag_pos_list)

                # 1. Declaring the entities
                result = corpus.call(
                    "store.declare_webentity_by_lrus",
                    list_lrus=prefixes,
                    name=name,
                    status=row[status_pos],
                    lruVariations=False,
                    startpages=startpages,
                    tags=tags,
                )

                webentity_id = result["result"]["id"]

                # 2. Setting the homepage
                try:
                    corpus.call(
                        "store.set_webentity_homepage",
                        webentity_id=webentity_id,
                        homepage=unquote(
                            homepage
                        ),  # TODO: drop unquote, linked to hyphe issue #447
                    )
                except HypheRequestFailError as e:
                    if "belong" in str(e):
                        loading_bar.print(
                            "Homepage error with %s %s" % (name, homepage)
                        )
                    else:
                        raise



# =============================================================================
# Minet CrowdTangle Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the CrowdTangle namespace.
#
from datetime import date, datetime, timedelta

DAY_DELTA = timedelta(days=1)


def day_range(end):
    day_delta = timedelta(days=1)

    start_date = date(*[int(i) for i in end.split("-")])
    current_date = date.today() + day_delta

    while start_date != current_date:
        end_date = current_date
        current_date -= day_delta

        yield current_date.isoformat(), end_date.isoformat()


def years_iter(start_date, end_date):
    current_year = int(end_date[:4])
    target_year = int(start_date[:4])

    current_start_date = str(current_year) + "-01-01T00:00:00"
    current_end_date = end_date

    while current_year > target_year:
        yield current_start_date, current_end_date

        current_year -= 1
        current_end_date = str(current_year) + "-12-31T23:59:59"
        current_start_date = str(current_year) + current_start_date[4:]

    yield start_date, current_end_date


def infer_end_date():
    return datetime.now().isoformat(timespec="seconds")


def get_last_day_of_month(year, month):
    d = 31

    while d > 27:
        try:
            attempt = date(int(year), int(month), d)
            return attempt.isoformat()[8:10]
        except ValueError:
            d -= 1

    raise TypeError("could not find last day of month")


def complement_date(d, bound):
    if len(d) == 4:
        d += "-"
        d += "01" if bound == "start" else "12"

    if len(d) == 7:
        d += "-"
        d += "01" if bound == "start" else get_last_day_of_month(d[:4], d[5:7])

    if len(d) == 10:
        d += "T00:00:00" if bound == "start" else "T23:59:59"

    return d


def step(request, url, item_key):
    data = request(url)
    items = None

    if item_key in data:
        items = data[item_key]

        if len(items) == 0:
            items = None

    # Extracting next link
    pagination = data["pagination"]
    next_page = pagination["nextPage"] if "nextPage" in pagination else None

    return items, next_page


def default_item_id_getter(item):
    return item["id"]


def make_paginated_iterator(
    url_forge, item_key, formatter, item_id_getter=default_item_id_getter
):
    def create_iterator(
        request,
        token,
        limit=None,
        raw=False,
        per_call=False,
        detailed=False,
        namespace=None,
        **kwargs,
    ):
        if namespace is not None:
            kwargs = vars(namespace)
        else:
            kwargs["token"] = token

        # Checking we have the necessary dates
        if kwargs.get("sort_by") == "date":
            # Inferring end date to be now, this will be important later
            if kwargs.get("end_date") is None:
                kwargs["end_date"] = infer_end_date()

        # Complementing dates
        if kwargs.get("start_date") is not None:
            kwargs["start_date"] = complement_date(kwargs["start_date"], "start")

        if kwargs.get("end_date") is not None:
            kwargs["end_date"] = complement_date(kwargs["end_date"], "end")

        N = 0
        C = 0
        last_items = set()

        has_limit = limit is not None

        # Chunking
        need_to_chunk = kwargs.get("sort_by") == "date"
        chunk_size = kwargs.get("chunk_size", 500)
        current_chunk_size = 0
        shifts = 0
        years = (
            years_iter(kwargs["start_date"], kwargs["end_date"])
            if need_to_chunk
            else None
        )

        def rotate_year():
            try:
                start_date, end_date = next(years)
                kwargs["start_date"] = start_date
                kwargs["end_date"] = end_date
            except StopIteration:
                return False

            return True

        if need_to_chunk:
            rotate_year()

        # Starting url
        url = url_forge(**kwargs)

        while True:
            C += 1

            items, next_url = step(request, url, item_key)

            # We have exhausted the available data
            if items is None:
                if need_to_chunk:
                    could_rotate = rotate_year()

                    if could_rotate:
                        url = url_forge(**kwargs)
                        continue

                break

            enough_to_stop = False
            n = 0

            acc = []

            for item in items:
                # Avoiding duplicating items due to race conditions
                if item_id_getter(item) in last_items:
                    continue

                current_date = item.get("date")

                n += 1
                N += 1
                current_chunk_size += 1

                if not raw:
                    item = formatter(item)

                acc.append(item)

                if has_limit and N >= limit:
                    enough_to_stop = True
                    break

            if per_call:
                if detailed:
                    details = None

                    if need_to_chunk:
                        details = {"date": current_date, "shifts": shifts}

                    yield details, acc
                else:
                    yield acc
            else:
                yield from acc

            if enough_to_stop:
                break

            # We need to track last items to avoid registering the same one twice
            last_items = set(item_id_getter(item) for item in items)

            # Paginating
            if next_url is None:
                if need_to_chunk:
                    could_rotate = rotate_year()

                    if could_rotate:
                        url = url_forge(**kwargs)
                        continue

                break

            # Handling chunking
            if need_to_chunk and current_chunk_size >= chunk_size:
                current_chunk_size = 0
                shifts += 1
                kwargs["end_date"] = items[-1]["date"].replace(" ", "T")
                url = url_forge(**kwargs)
                continue

            url = next_url

    return create_iterator



# =============================================================================
# Minet Instagram Utils
# =============================================================================
#
# Miscellaneous helper functions used by the minet.instagram namespace.
#

import re

HASHTAGS_RE = re.compile(r"[^\w#]+")
HANDLES_RE = re.compile(r"[^\w\.@]+")


def extract_hashtags(text):
    return sorted(
        set(r.lstrip("#").lower() for r in HASHTAGS_RE.split(text) if r.startswith("#"))
    )


def extract_handles(text):
    return sorted(
        set(r.lstrip("@").lower() for r in HANDLES_RE.split(text) if r.startswith("@"))
    )


def short_code_to_url(short_code):
    return "https://www.instagram.com/p/{}/".format(short_code)



# =============================================================================
# Minet Mediacloud Search
# =============================================================================
#
# Function related to stories searching.
#
from urllib.parse import quote_plus

from minet.web import request
from minet.mediacloud.constants import MEDIACLOUD_API_BASE_URL, MEDIACLOUD_DEFAULT_BATCH
from minet.mediacloud.exceptions import MediacloudServerError
from minet.mediacloud.types import MediacloudStory
from minet.mediacloud.utils import get_last_processed_stories_id


def create_plural_query_component(field, values):
    query = " AND ("
    query += " OR ".join(("%s:%s" % (field, str(value))) for value in values)
    query += ")"

    return query


def pad_date(string):
    if len(string) == 4:
        return string + "-01-01T00:00:00Z"

    if len(string) == 7:
        return string + "-01T00:00:00Z"

    if "T" not in string:
        return string + "T00:00:00Z"

    return string


def query_additions(
    query,
    collections=None,
    medias=None,
    publish_day=None,
    publish_month=None,
    publish_year=None,
):
    if collections is not None:
        query += create_plural_query_component("tags_id_media", collections)

    if medias is not None:
        query += create_plural_query_component("media_id", medias)

    if publish_day is not None:
        query += ' AND publish_day:"%s"' % pad_date(publish_day)

    if publish_month is not None:
        query += ' AND publish_month:"%s"' % pad_date(publish_month)

    if publish_year is not None:
        query += ' AND publish_year:"%s"' % pad_date(publish_year)

    # NOTE: range queries can be made with [from_date TO to_date] but must appear in filter

    return query


def url_forge(
    token,
    query,
    filter_query=None,
    collections=None,
    medias=None,
    publish_day=None,
    publish_month=None,
    publish_year=None,
    count=False,
    last_processed_stories_id=None,
):
    url = "%s/stories_public/%s?key=%s" % (
        MEDIACLOUD_API_BASE_URL,
        "count" if count else "list",
        token,
    )

    query = query_additions(
        query,
        collections=collections,
        medias=medias,
        publish_day=publish_day,
        publish_month=publish_month,
        publish_year=publish_year,
    )

    url += "&q=%s" % quote_plus(query)

    if filter_query is not None:
        url += "&fq=%s" % quote_plus(filter_query)

    if not count:
        url += "&rows=%i" % MEDIACLOUD_DEFAULT_BATCH

        if last_processed_stories_id is not None:
            url += "&last_processed_stories_id=%i" % last_processed_stories_id

    return url


def mediacloud_search(
    pool_manager,
    token,
    query,
    filter_query=None,
    count=False,
    collections=None,
    medias=None,
    publish_day=None,
    publish_month=None,
    publish_year=None,
    raw=False,
):
    def generator():
        last_processed_stories_id = None

        while True:
            url = url_forge(
                token,
                query,
                filter_query=filter_query,
                collections=collections,
                medias=medias,
                publish_day=publish_day,
                publish_month=publish_month,
                publish_year=publish_year,
                count=count,
                last_processed_stories_id=last_processed_stories_id,
            )

            response = request(url, pool_manager=pool_manager)
            data = response.json()

            if response.status >= 500:
                raise MediacloudServerError(server_error=data.get("error"))

            if count:
                yield data["count"]
                return

            for story in data:
                if not raw:
                    story = MediacloudStory.from_payload(story)

                yield story

            last_processed_stories_id = get_last_processed_stories_id(data)

            if last_processed_stories_id is None:
                return

    if count:
        return next(generator())

    return generator()



# =============================================================================
# Minet Instagram Constants
# =============================================================================
#
# General constants used throughout the Instagram functions.
#
from urllib3 import Timeout

INSTAGRAM_URL = "https://www.instagram.com"
INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
INSTAGRAM_MEDIA_TYPE = {1: "GraphImage", 2: "GraphVideo", 8: "GraphSidecar"}
INSTAGRAM_DEFAULT_THROTTLE = 30.0
INSTAGRAM_MAX_RANDOM_ADDENDUM = 10.0
INSTAGRAM_NB_REQUEST_LITTLE_WAIT = 10
INSTAGRAM_DEFAULT_THROTTLE_LITTLE_WAIT = 300.0
INSTAGRAM_MAX_RANDOM_ADDENDUM_LITTLE_WAIT = 300.0
INSTAGRAM_NB_REQUEST_BIG_WAIT = 2900
INSTAGRAM_DEFAULT_THROTTLE_BIG_WAIT = 1800.0
INSTAGRAM_MAX_RANDOM_ADDENDUM_BIG_WAIT = 1800.0
INSTAGRAM_MIN_TIME_RETRYER = 600.0



# =============================================================================
# Minet Instagram User-posts CLI Action
# =============================================================================
#
# Logic of the `instagram user-posts` action.
#
from itertools import islice

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.instagram.utils import with_instagram_fatal_errors
from minet.instagram import InstagramAPIScraper
from minet.instagram.types import InstagramPost
from minet.instagram.exceptions import (
    InstagramNoPublicationError,
    InstagramPrivateOrNonExistentAccountError,
    InstagramInvalidTargetError,
)


@with_instagram_fatal_errors
@with_enricher_and_loading_bar(
    headers=InstagramPost,
    title="Scraping posts",
    unit="users",
    nested=True,
    sub_unit="posts",
)
def action(cli_args, enricher, loading_bar):
    client = InstagramAPIScraper(cookie=cli_args.cookie)

    for i, row, user in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step(user):
            try:
                generator = client.user_posts(user)

                if cli_args.limit:
                    generator = islice(generator, cli_args.limit)

                for post in generator:
                    enricher.writerow(row, post)
                    loading_bar.nested_advance()

            except InstagramInvalidTargetError:
                loading_bar.print(
                    "Given user (line %i) is probably not an Instagram user: %s"
                    % (i, user)
                )

            except InstagramPrivateOrNonExistentAccountError:
                loading_bar.print(
                    "Given user (line %i) is probably a private Instagram account or is not an Instagram user: %s"
                    % (i, user)
                )

            except InstagramNoPublicationError:
                loading_bar.print(
                    "Given user (line %i) has probably no publication: %s" % (i, user)
                )



# =============================================================================
# Minet Mediacloud Medias CLI Action
# =============================================================================
#
# Logic of the `mc medias` action.
#
import casanova

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.mediacloud.utils import with_mediacloud_fatal_errors
from minet.mediacloud import MediacloudAPIClient
from minet.mediacloud.types import MediacloudFeed, MediacloudMedia


def get_headers(cli_args):
    headers = MediacloudMedia.fieldnames()[1:]

    if cli_args.feeds is not None:
        headers.append("feeds")

    return headers


@with_mediacloud_fatal_errors
@with_enricher_and_loading_bar(
    headers=get_headers, title="Fetching medias", unit="medias"
)
def action(cli_args, enricher, loading_bar):
    feeds_writer = None

    if cli_args.feeds:
        feeds_writer = casanova.writer(cli_args.feeds, fieldnames=MediacloudFeed)

    client = MediacloudAPIClient(cli_args.token)

    for row, media_id in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            result = client.media(media_id)
            result = result.as_csv_row()[1:]

            if cli_args.feeds:
                assert feeds_writer is not None

                feeds = client.feeds(media_id)

                enricher.writerow(row, result + [len(feeds)])

                for feed in feeds:
                    feeds_writer.writerow(feed.as_csv_row())
            else:
                enricher.writerow(row, result)



from typing import List

from dataclasses import dataclass
from casanova import TabularRecord


@dataclass
class TelegramChannelInfos(TabularRecord):
    title: str
    name: str
    link: str
    img: str
    description: str
    nb_subscribers: str
    nb_photos: str
    nb_videos: str
    nb_links: str


@dataclass
class TelegramChannelMessages(TabularRecord):
    link_to_message: str
    could_be_displayed: bool
    user: str
    user_link: str
    user_img: str
    text: str
    nb_hashtags: int
    hashtags: List[str]
    is_reply_img: bool
    is_reply_user: bool
    is_reply_text: bool
    stickers: List[str]
    nb_photos: int
    photos: List[str]
    nb_videos: int
    videos: List[str]
    videos_times: List[str]
    nb_links: int
    links: List[str]
    link_img: str
    link_site: str
    link_title: str
    link_description: str
    views: int
    datetime: str
    edited: bool



# =============================================================================
# Minet Youtube Captions CLI Action
# =============================================================================
#
# Action retrieving the captions of YouTube videos using the API.
#
from minet.youtube import YouTubeScraper
from minet.youtube.types import YouTubeCaptionTrack, YouTubeCaptionLine
from minet.youtube.exceptions import YouTubeInvalidVideoTargetError

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.loading_bar import LoadingBar


def get_headers(cli_args):
    headers = YouTubeCaptionTrack.fieldnames(prefix="caption_track_")

    if cli_args.collapse:
        return headers + ["caption_track_text"]

    return headers + YouTubeCaptionLine.fieldnames(prefix="caption_line_")


@with_enricher_and_loading_bar(
    headers=get_headers, title="Retrieving captions", unit="videos"
)
def action(cli_args, enricher, loading_bar: LoadingBar):
    scraper = YouTubeScraper()

    lang_warning_printed = False

    for row, video in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            try:
                result = scraper.get_video_captions(video, langs=cli_args.lang)
            except YouTubeInvalidVideoTargetError:
                loading_bar.inc_stat("not-a-video", style="error")
                continue

            if result is None:
                if not lang_warning_printed:
                    lang_warning_printed = True
                    loading_bar.warning(
                        "Did you forget to pass the correct value to [dim]--lang[/dim]?"
                    )

                loading_bar.inc_stat("not-found", style="error")
                continue

            track, lines = result

            if cli_args.collapse:
                enricher.writerow(row, track, ["\n".join(line.text for line in lines)])
            else:
                for line in lines:
                    enricher.writerow(row, track, line)



# =============================================================================
# Minet Facebook Post Id From Url
# =============================================================================
#
# Helper used to retrieved a full facebook post id from the given post url.
#
import re
import json
from urllib.parse import urlsplit, parse_qsl, urljoin
from ural.facebook import (
    convert_facebook_url_to_mobile,
    parse_facebook_url,
    FacebookPost,
)

from minet.rate_limiting import rate_limited_from_state
from minet.web import request
from minet.facebook.constants import (
    FACEBOOK_URL,
    FACEBOOK_MOBILE_URL,
    FACEBOOK_MOBILE_RATE_LIMITER_STATE,
)

PAGE_ID_PATTERN = re.compile(r"&amp;rid=(\d+)&amp;")
GROUP_ID_PATTERN = re.compile(r"fb://group/(\d+)")


@rate_limited_from_state(FACEBOOK_MOBILE_RATE_LIMITER_STATE)
def page_id_from_handle(handle):
    url = urljoin(FACEBOOK_MOBILE_URL, handle)

    response = request(url, headers={"User-Agent": "curl/7.68.0"})

    if response.status >= 400:
        return None

    m = PAGE_ID_PATTERN.search(response.text())

    if m is None:
        return None

    return m.group(1)


@rate_limited_from_state(FACEBOOK_MOBILE_RATE_LIMITER_STATE)
def group_id_from_handle(handle):
    url = urljoin(FACEBOOK_MOBILE_URL, "groups/%s" % handle)

    response = request(url, headers={"User-Agent": "curl/7.68.0"})

    if response.status >= 400:
        return None

    m = GROUP_ID_PATTERN.search(response.text())

    if m is None:
        return None

    return m.group(1)


@rate_limited_from_state(FACEBOOK_MOBILE_RATE_LIMITER_STATE)
def scrape_post_id(post_url):
    post_mobile_url = convert_facebook_url_to_mobile(post_url)

    soup = request(post_mobile_url).soup()

    root_element = soup.select_one("#m_story_permalink_view [data-ft]")

    if root_element is None:
        # Is this a photo post?
        next_link = soup.select_one('[href^="/photo.php"]')

        if next_link is None:
            return

        href = next_link.get("href")

        if not href:
            return

        link = urljoin(FACEBOOK_URL, href)
        query = urlsplit(link).query

        if not query:
            return

        query = dict(parse_qsl(query))

        return "%s_%s" % (query["id"], query["fbid"])

    data = root_element.get("data-ft")

    if data is None:
        return

    try:
        data = json.loads(data)
    except json.JSONDecodeError:
        return

    content_owner_id_new = data.get("content_owner_id_new") or data.get("page_id")
    mf_story_key = data.get("mf_story_key")

    if content_owner_id_new is None or mf_story_key is None:
        return

    return "%s_%s" % (content_owner_id_new, mf_story_key)


# TODO: could easily cache some retrieved handles...
def post_id_from_url(post_url):
    parsed = parse_facebook_url(post_url)

    if not isinstance(parsed, FacebookPost):
        return

    # NOTE: as hinted in #768, we cannot figure pfbids
    if parsed.id.startswith("pfbid"):
        return scrape_post_id(post_url)

    if parsed.full_id is not None:
        return parsed.full_id

    if parsed.parent_handle is not None:
        parent_id = page_id_from_handle(parsed.parent_handle)

        if parent_id is not None:
            return "%s_%s" % (parent_id, parsed.id)

    elif parsed.group_handle is not None:
        group_id = group_id_from_handle(parsed.group_handle)

        if group_id is not None:
            return "%s_%s" % (group_id, parsed.id)

    return scrape_post_id(post_url)



# =============================================================================
# Minet CrowdTangle Leaderboard CLI Action
# =============================================================================
#
# Logic of the `ct leaderboard` action.
#
from minet.crowdtangle.types import CrowdTangleLeaderboard
from minet.cli.crowdtangle.utils import make_paginated_action


action = make_paginated_action(
    method_name="leaderboard",
    item_name="accounts",
    csv_headers=CrowdTangleLeaderboard.fieldnames(),
)



# =============================================================================
# Minet Facebook User Infos CLI Action
# =============================================================================
#
# Logic of the `fb user-infos` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.loading_bar import LoadingBar
from minet.cli.facebook.utils import with_facebook_fatal_errors
from minet.facebook import FacebookMobileScraper
from minet.facebook.types import MobileFacebookUserInfo


@with_facebook_fatal_errors
@with_enricher_and_loading_bar(
    headers=MobileFacebookUserInfo, title="Scraping user infos", unit="users"
)
def action(cli_args, enricher, loading_bar: LoadingBar):
    scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)

    for row, user_url in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            try:
                user_infos = scraper.user_infos(user_url)
                enricher.writerow(row, user_infos)
            except TypeError:
                enricher.writerow(row)
                loading_bar.inc_stat("invalid-url", style="error")



from minet.cli.argparse import command, InputAction

URL_EXTRACT_COMMAND = command(
    "url-extract",
    "minet.cli.url_extract.url_extract",
    title="Minet Url Extract Command",
    description="""
        Extract urls from a CSV column containing either raw text or raw
        HTML.
    """,
    epilog="""
        Examples:

        . Extracting urls from a text column:
            $ minet url-extract text -i posts.csv > urls.csv

        . Extracting urls from a html column:
            $ minet url-extract html --from html -i posts.csv > urls.csv
    """,
    select=True,
    total=True,
    arguments=[
        {"name": "column", "help": "Name of the column containing text or html."},
        {"name": "input", "help": "Target CSV file.", "action": InputAction},
        {"flag": "--base-url", "help": "Base url used to resolve relative urls."},
        {
            "flag": "--from",
            "help": "Extract urls from which kind of source?",
            "choices": ["text", "html"],
            "default": "text",
        },
    ],
)



from minet.cli.argparse import command, InputAction

URL_JOIN_COMMAND = command(
    "url-join",
    "minet.cli.url_join.url_join",
    title="Minet Url Join Command",
    description="""
        Join two CSV files by matching them on columns containing urls. It
        works by indexing the first file's urls in a specialized
        URL trie to match them with the second file's urls.
    """,
    epilog="""
        Examples:

        . Joining two files:
            $ minet url-join url webentities.csv post_url posts.csv > joined.csv

        . Adding a prefix to first file's column names:
            $ minet url-join -p w_ url webentities.csv post_url posts.csv > joined.csv

        . Keeping only some columns from first file:
            $ minet url-join url webentities.csv post_url posts.csv -s url,id > joined.csv

        . Reading one of the files from stdin:
            $ xan select url webentities.csv | minet url-join url - post_url posts.csv > joined.csv
    """,
    select=True,
    arguments=[
        {
            "name": "column1",
            "help": "Name of the column containing urls in the indexed file.",
        },
        {
            "name": "input1",
            "help": "Path to the file to index. Will understand `-` as stdin.",
            "action": InputAction,
            "nargs": None,
        },
        {
            "name": "column2",
            "help": "Name of the column containing urls in the second file.",
        },
        {
            "name": "input2",
            "help": "Path to the second file. Will understand `-` as stdin.",
            "action": InputAction,
            "nargs": None,
        },
        {
            "flags": ["-p", "--match-column-prefix"],
            "help": "Optional prefix to add to the first file's column names to avoid conflicts.",
            "default": "",
        },
        {"flag": "--separator", "help": "Split indexed url column by a separator?"},
    ],
)
